class Token:
    """
    Container class for token types.

    """

    class _Base:
        indents = False
        dedents = False
        ignore = False

        def __init__(self, text, kind="", offset=0):
            self.text = text
            self.kind = kind
            self.offset = offset

        def __str__(self):
            return self.text

        def __repr__(self):
            return f"({self.__class__.__name__}:{repr(self.text)})"

    class Text(_Base):
        pass

    class Ignore(_Base):
        ignore = True

    class Open(_Base):
        indents = True

    class Close(_Base):
        dedents = True

    class OpenAndClose(_Base):
        indents = True
        dedents = True



import argparse
import sys

from . import modes


def verify_changed(source, result):
    """
    Verify that the source is either exactly equal to the result or
    that the result has only changed by added or removed whitespace.

    """
    output_lines = result.split("\n")
    changed = False
    for line_nr, line in enumerate(source.split("\n")):
        if line != output_lines[line_nr]:
            changed = True
        if line.strip() != output_lines[line_nr].strip():
            raise IndentationError("Non-whitespace changes detected. Core dumped.")

    return changed


def main():
    """
    Entrypoint for all 4 command-line tools. Typical usage:

        $ djhtml -i file1.html file2.html

    """
    Mode = modes.DjHTML
    if sys.argv[0].endswith("djtxt"):
        Mode = modes.DjTXT
    if sys.argv[0].endswith("djcss"):
        Mode = modes.DjCSS
    if sys.argv[0].endswith("djjs"):
        Mode = modes.DjJS

    changed_files = 0
    unchanged_files = 0
    problematic_files = 0

    parser = argparse.ArgumentParser(
        description=(
            "DjHTML is a fully automatic template indenter that works with mixed"
            " HTML/CSS/Javascript templates that contain Django or Jinja2 template"
            " tags. It works similar to other code-formatting tools such as Black and"
            " interoperates nicely with pre-commit. More information at"
            " https://github.com/rtts/djhtml"
        ),
    )
    parser.add_argument(
        "-i", "--in-place", action="store_true", help="modify files in-place"
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="be quiet")
    parser.add_argument(
        "-t", "--tabwidth", metavar="N", type=int, default=4, help="tab width"
    )
    parser.add_argument(
        "-o",
        "--output-file",
        metavar="filename",
        default="-",
        help="output filename",
    )
    parser.add_argument(
        "input_filenames",
        metavar="filenames",
        nargs="*",
        default=["-"],
        help="input filenames",
    )
    parser.add_argument("-d", "--debug", action="store_true", help=argparse.SUPPRESS)
    args = parser.parse_args()

    if args.in_place and "-" in args.input_filenames:
        sys.exit("I’m sorry Dave, I’m afraid I can’t do that")

    if len(args.input_filenames) > 1 and not args.in_place:
        sys.exit("Will not modify files in-place without -i option")

    for input_filename in args.input_filenames:

        # Read input file
        try:
            input_file = (
                sys.stdin if input_filename == "-" else open(input_filename, "r")
            )
            source = input_file.read()
        except Exception as e:
            problematic_files += 1
            if not args.quiet:
                print(f"Error opening {input_filename}: {e}", file=sys.stderr)
            continue

        # Indent input file
        try:
            if args.debug:
                print(Mode(source).debug())
                sys.exit()
            result = Mode(source).indent(args.tabwidth)
        except SyntaxError as e:
            problematic_files += 1
            if not args.quiet:
                print(
                    f"Syntax error in {input_file.name}:"
                    f" {str(e) or e.__class__.__name__}",
                    file=sys.stderr,
                )
            continue
        except Exception:
            print(
                f"\nFatal error while processing {input_file.name}\n\n"
                "    If you have time and are using the latest version, we\n"
                "    would very much appreciate if you opened an issue on\n"
                "    https://github.com/rtts/djhtml/issues\n",
                file=sys.stderr,
            )
            raise
        finally:
            input_file.close()

        changed = verify_changed(source, result)

        # Print to stdout and exit
        if not args.in_place and args.output_file == "-":
            if not args.quiet:
                print(result, end="")
            sys.exit(0)  # YOLO

        # Write output file
        if changed:
            output_filename = input_file.name if args.in_place else args.output_file
            try:
                output_file = open(output_filename, "w")
                output_file.write(result)
                output_file.close()
                changed_files += 1
            except Exception as e:
                problematic_files += 1
                if not args.quiet:
                    print(f"Error writing {output_filename}: {e}", file=sys.stderr)
                continue
            if not args.quiet:
                print(
                    f"reindented {output_file.name}",
                    file=sys.stderr,
                )
        else:
            unchanged_files += 1

    # Print final summary
    if not args.quiet:
        s = "s" if changed_files != 1 else ""
        have = "have" if s else "has"
        print(
            f"{changed_files} template{s} {have} been reindented.",
            file=sys.stderr,
        )
        if unchanged_files:
            s = "s" if unchanged_files != 1 else ""
            were = "were" if s else "was"
            print(
                f"{unchanged_files} template{s} {were} already perfect!",
                file=sys.stderr,
            )
        if problematic_files:
            s = "s" if problematic_files != 1 else ""
            print(
                f"{problematic_files} template{s} could not be processed due to an"
                " error.",
                file=sys.stderr,
            )

    sys.exit(problematic_files)


if __name__ == "__main__":
    main()



class Line:
    """
    A single output line including the final newline.

    """

    def __init__(self, nr=1):
        self.nr = nr
        self.tokens = []
        self.level = 0

    def append(self, token):
        """
        Append tokens to the line.

        """
        token.line_nr = self.nr
        self.tokens.append(token)

    @property
    def text(self):
        """
        The unindented text of this line without leading/trailing spaces.

        """
        return "".join([str(token) for token in self.tokens]).strip()

    def indent(self, tabwidth):
        """
        The final, indented text of this line. Make sure to set the level
        and optionally offset before calling this method.

        """
        if self.tokens:
            if self.tokens[0].ignore:
                return "".join([str(token) for token in self.tokens]) + "\n"
            elif self.text:
                offset = self.tokens[0].offset * tabwidth
                spaces = tabwidth * self.level + offset
                return " " * spaces + self.text + "\n"
        return "\n"

    def __repr__(self):
        return repr(self.tokens)

    def __bool__(self):
        return bool(self.tokens and self.text)

    def __next__(self):
        return Line(nr=self.nr + 1)



import re

from .lines import Line
from .tokens import Token


class DjTXT:
    """
    Mode for indenting text files that contain Django template tags.
    Also serves as the base class for the other modes:

    - DjHTML
    - DjCSS
    - DjJS

    """

    RAW_TOKENS = [
        r"\n",
        r"{%.*?%}",
        r"{#.*?#}",
    ]

    DJANGO_OPENING_AND_CLOSING_TAGS = [
        "elif",
        "else",
        "empty",
        "plural",
    ]

    def __init__(self, source="", return_mode=None):
        self.source = source
        self.return_mode = return_mode or self
        self.token_re = compile_re(self.RAW_TOKENS)

    def indent(self, tabwidth):
        """
        Return the indented text as a single string.

        """
        self.tokenize()
        self.parse()
        return "".join([line.indent(tabwidth) for line in self.lines])

    def parse(self):
        """
        You found the top-secret indenting algorithm!

        """
        stack = []
        for line in self.lines:
            first_token = True
            for token in line.tokens:
                opening_token = None

                # When a dedenting token is found, match it with the
                # token at the top of the stack.
                if token.dedents:
                    try:
                        if stack[-1].kind == token.kind:
                            opening_token = stack.pop()
                        elif token.kind == "django":
                            opening_token = stack.pop()
                            while opening_token.kind != token.kind:
                                opening_token = stack.pop()
                        elif first_token:
                            # This closing token could not be matched.
                            # Instead of erroring out, set the line level
                            # to what it would have been with a
                            # regular text token.
                            line.level = stack[-1].level + 1
                    except IndexError:
                        line.level = 0

                    # If this dedenting token is the first in line,
                    # set the line level to the line level of the
                    # corresponding opening token.
                    if first_token and opening_token:
                        line.level = opening_token.level

                # If the first token is not a dedenting token, the
                # line level will be one higher than that of the token
                # at the top of the stack.
                elif first_token:
                    line.level = stack[-1].level + 1 if stack else 0

                # Push indenting tokens onto the stack. Note that some
                # tokens can be both indenting and dedenting (e.g.,
                # ``{% else %}``), hence the if instead of elif.
                if token.indents:
                    token.level = opening_token.level if opening_token else line.level
                    stack.append(token)

                # Subsequent tokens have no effect on the line level
                # (but tokens with only spaces don't count).
                if token.text.strip():
                    first_token = False

    def tokenize(self):
        """
        Split the source text into tokens and place them on lines.

        """
        self.lines = []
        line = Line()
        mode = self
        src = self.source

        while True:
            try:
                # Split the source at the first occurence of one of
                # the current mode's raw tokens.
                head, raw_token, tail = mode.token_re.split(src, maxsplit=1)

            except ValueError:
                # We've reached the final line!
                if src:
                    line.append(mode.create_token(src, ""))
                if line:
                    self.lines.append(line)
                break

            if head:
                # Create a token from the head. This will always be a
                # text token (and the next mode will always be the
                # current mode), but we don't assume that here.
                line.append(mode.create_token(head, raw_token + tail))
                mode = next(mode)

            if raw_token == "\n":
                self.lines.append(line)
                line = next(line)

            else:
                # Ask the mode to create a token and to provide a new
                # mode for the next iteration of the loop.
                line.append(mode.create_token(raw_token, tail))
                mode = next(mode)

            # Set the new source to the old tail for the next iteration.
            src = tail

    def create_token(self, raw_token, src):
        """
        Given a raw token string, return a single token (and internally
        set the next mode).

        """
        kind = "django"
        self.next_mode = self
        token = Token.Text(raw_token)

        tag = re.match(r"{% *(\w+).*?%}", raw_token)
        comment = None if tag else re.match(r"{# *(\w+:\w+).*?#}", raw_token)
        if tag:
            name = tag.group(1)
            if name == "comment":
                token = Token.Open(raw_token, kind)
                self.next_mode = Comment(r"\{% *endcomment.*?%\}", self, kind)
            elif re.search(f"{{% *end{name}.*?%}}", src):
                token = Token.Open(raw_token, kind)
            elif name in self.DJANGO_OPENING_AND_CLOSING_TAGS:
                token = Token.OpenAndClose(raw_token, kind)
            elif name.startswith("end"):
                token = Token.Close(raw_token, kind)

        elif comment:
            name = comment.group(1)
            if name == "fmt:off":
                token = Token.Open(raw_token, kind)
                self.next_mode = Comment(r"\{% *fmt:on.*?%\}", self, kind)

        return token

    def debug(self):
        self.tokenize()
        return "\n".join(
            [" ".join([repr(token) for token in line.tokens]) for line in self.lines]
        )

    def __next__(self):
        return self.next_mode


class DjHTML(DjTXT):
    """
    This mode is the entrypoint of DjHTML. Usage:

    >>> DjHTML(input_string).indent(tabwidth=4)

    """

    RAW_TOKENS = DjTXT.RAW_TOKENS + [
        r"<pre.*?>",
        r"</.*?>",
        r"<!--",
        r"<",
    ]

    IGNORE_TAGS = [
        "area",
        "base",
        "br",
        "col",
        "command",
        "embed",
        "hr",
        "img",
        "input",
        "keygen",
        "link",
        "meta",
        "param",
        "source",
        "track",
        "wbr",
    ]

    def create_token(self, raw_token, src):
        kind = "html"
        self.next_mode = self

        if raw_token == "<":
            tag = re.match(r"(\w+)[^:]", src)
            if tag:
                token = Token.Open(raw_token, kind)
                self.next_mode = InsideHTMLTag(tag[1], self)
            else:
                token = Token.Text(raw_token)
            return token

        if raw_token == "<!--":
            self.next_mode = Comment("-->", self, kind)
            return Token.Open(raw_token, kind)

        if re.match("<pre.*?>", raw_token):
            self.next_mode = Comment("</pre>", self, kind)
            return Token.Open(raw_token, kind)

        if raw_token.startswith("</"):
            tagname = re.search(r"\w+", raw_token)
            if tagname:
                if tagname[0].lower() in self.IGNORE_TAGS:
                    return Token.Text(raw_token)
            return Token.Close(raw_token, kind)

        return super().create_token(raw_token, src)


class DjCSS(DjTXT):
    """
    Mode for indenting CSS.

    """

    RAW_TOKENS = DjTXT.RAW_TOKENS + [
        r"</style>",
        r"{",
        r"}",
        r"/\*",
    ]

    def create_token(self, raw_token, src):
        kind = "css"
        self.next_mode = self

        if raw_token == "{":
            return Token.Open(raw_token, kind)
        if raw_token == "}":
            return Token.Close(raw_token, kind)
        if raw_token == "/*":
            self.next_mode = Comment(r"\*/", self, kind)
            return Token.Open(raw_token, kind)
        if raw_token == "</style>":
            self.next_mode = self.return_mode
            return Token.Close(raw_token, "html")

        return super().create_token(raw_token, src)


class DjJS(DjTXT):
    """
    Mode for indenting Javascript.

    """

    RAW_TOKENS = DjTXT.RAW_TOKENS + [
        r"</script>",
        r'".*?"',
        r"'.*?'",
        r"`.*?`",
        r"`",
        r"[\{\[\(\)\]\}]",
        r"//.*",
        r"/\*",
    ]

    def create_token(self, raw_token, src):
        kind = "javascript"
        self.next_mode = self

        if raw_token in "{[(":
            return Token.Open(raw_token, kind)
        if raw_token in ")]}":
            return Token.Close(raw_token, kind)
        if raw_token == "`":
            self.next_mode = Comment("`", self, kind)
            return Token.Open(raw_token, kind)
        if raw_token == "/*":
            self.next_mode = Comment(r"\*/", self, kind)
            return Token.Open(raw_token, kind)
        if raw_token.lstrip().startswith("."):
            return Token.Text(raw_token, offset=1)
        if raw_token == "</script>":
            self.next_mode = self.return_mode
            return Token.Close(raw_token, "html")

        return super().create_token(raw_token, src)


# The following are "special" modes with different constructors.


class Comment(DjTXT):
    """
    Mode to create ignore tokens until an end tag is encountered.

    """

    def __init__(self, endtag, return_mode, kind):
        self.endtag = endtag
        self.return_mode = return_mode
        self.kind = kind
        self.token_re = compile_re([r"\n", endtag])

    def create_token(self, raw_token, src):
        self.next_mode = self
        if re.match(self.endtag, raw_token):
            self.next_mode = self.return_mode
            return Token.Close(raw_token, self.kind)
        return Token.Ignore(raw_token)


class InsideHTMLTag(DjTXT):
    """
    Welcome to the wondrous world between "<" and ">".

    """

    RAW_TOKENS = DjTXT.RAW_TOKENS + [r"/?>"]

    def __init__(self, tagname, return_mode):
        self.tagname = tagname
        self.return_mode = return_mode
        self.token_re = compile_re(self.RAW_TOKENS)

    def create_token(self, raw_token, src):
        kind = "html"
        self.next_mode = self

        if raw_token == "/>":
            self.next_mode = self.return_mode
            return Token.Close(raw_token, kind)
        elif raw_token == ">":
            if self.tagname.lower() in DjHTML.IGNORE_TAGS:
                self.next_mode = self.return_mode
                return Token.Close(raw_token, kind)
            else:
                if self.tagname == "style":
                    self.next_mode = DjCSS(return_mode=self.return_mode)
                elif self.tagname == "script":
                    self.next_mode = DjJS(return_mode=self.return_mode)
                else:
                    self.next_mode = self.return_mode
                return Token.OpenAndClose(raw_token, kind)
        elif "text/template" in raw_token:
            self.tagname = ""

        return super().create_token(raw_token, src)


def compile_re(raw_tokens):
    return re.compile("(" + "|".join(raw_tokens) + ")")
